import numpy as np 
import sympy
from sympy.matrices import Matrix
def rxRateFun(c, rc):
	"""
	Computes the reaction rate vector for a given concentration

	INPUT
	c: numpy array. State space vector which the AR will be computed in (usually concentrations or
		concentrations and residence time space)
	rc: dictionary. Information about the rate constants
	OUTPUT
	r: numpy array. The rate vector
	"""
	# Unpack the rate vector
	rmax = rc['rmax'] # mol/(L h)
	Kma = rc['Kma'] # [mol/L]
	Kmb = rc['Kmb'] # [mol/L]
	Kia = rc['Kia'] # [mol/L]
	cf = np.asmatrix(rc['cf']).T # [g/L]
	cfAll = np.asmatrix(rc['cfAll']).T # [g/L], np.array([cbuoh, chbu, cbub, cs, tau])
	isub = rc['isub'] # indices in vectors
	ihbu = rc['ihbu']
	ibuoh = rc['ibuoh']
	ibub = rc['ibub']
	MW = rc['MW'] #[g/mol], np.array

	#make c = [cbub, tau] a matrix in order to multiply
	ci = np.asmatrix(c).T # independent concentrations
	# mass balance matrix
	A = np.matrix([[-MW[ibuoh]/MW[ibub], 0], #buoh
		[-MW[ihbu]/MW[ibub], 0], # hbu
		[1, 0], # bub
		[0, 0], # s
		[0, 1]]) # tau

	cAll = cfAll + A*(ci - cf)

	# compute the rate of reaction, in molar form. Rewrite chbu = ca and 
	# cbuoh = cb, where ca,cb are in mol/L
	ca = cAll[ihbu, 0]/MW[ihbu]
	cb = cAll[ibuoh, 0]/MW[ibuoh]

	rateOfRx = (rmax * ca * cb) / (Kia*Kmb + Kma*cb + Kmb*ca + ca*cb)

	r = np.zeros(np.shape(c))
	r[0] = MW[ibub] * rateOfRx # [g/L h], dbub/dt
	r[1] = 1 # [-] dt/dt
	return r

def solveMoleBalance(cin, rc):
	"""
	Solve the mole/mass balance for the system (determine the dependent concentrations given the 
	independent concentrations)
	INPUT
	cin: numpy array. State space vector which the AR will be computed in (usually concentrations or
		concentrations and residence time space). The independent variables
	rc: dictionary. Information about the rate constants
	OUTPUT
	cout: numpy array. State space vector for all independent and dependent variabels (independent 
		concentrations + residence time (if this is variabel in the specific AR) AND dependent
		concentrations
	"""

	# Unpack the rate vector
	# Unpack the rate vector
	rmax = rc['rmax'] # mol/(L h)
	Kma = rc['Kma'] # [mol/L]
	Kmb = rc['Kmb'] # [mol/L]
	Kia = rc['Kia'] # [mol/L]
	cf = np.asmatrix(rc['cf']).T # [g/L]
	cfAll = np.asmatrix(rc['cfAll']).T # [g/L], np.array([cbuoh, chbu, cbub, cs, tau])
	isub = rc['isub'] # indices in vectors
	ihbu = rc['ihbu']
	ibuoh = rc['ibuoh']
	ibub = rc['ibub']
	MW = rc['MW'] #[g/mol], np.array

	#make c = [cbub, tau] a matrix in order to multiply
	# ci = np.asmatrix(c).T # independent concentrations
	# mass balance matrix
	A = np.matrix([[-MW[ibuoh]/MW[ibub], 0], #buoh
		[-MW[ihbu]/MW[ibub], 0], # hbu
		[1, 0], # bub
		[0, 0], # s
		[0, 1]]) # tau

	# cAll = cfAll + A*(ci - cf)
	# write delta-concentrations = c - cf for the independent species
	cAll = np.zeros((cin.shape[0], len(cfAll)))
	for i in range(len(cin[:, 0])):
		cInd = np.asmatrix(cin[i, :]).T
		# ci = np.matrix([[cs], [cbuoh]])
		dci = cInd - cf #change of independent concentrations
		cAlli = cfAll + A*dci #mass balance for dependent and independent species
		cAll[i, :] = np.asarray(cAlli)[:, 0]
	# cout = np.column_stack((cAll[:, ix], cAll[:, isub], cAll[:, ihbu], cAll[:, ibuoh], cAll[:, ietoh]))
	return cAll

#def rxRateSympy(rc):
	"""
	Returns the reaction rate vector with symbolic expressions. Also return the expressions themselves
	INPUT
	rc: dictionary. Information about the rate constants
	OUTPUT
	r: sympy Matrix. The rate vector with symbolic expressions
	sym: list of sympy.symbols
	"""
	muMax = rc['muMax']
	Ks = rc['Ks']
	kd = rc['kd']
	Yxs = rc['Yxs']
	Ypx = rc['Ypx']
	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
	mu = (muMax * cs) / (Ks + cs)
	r = Matrix([mu*cx - kd*cx, #rx
		-(mu/Yxs)*cx, #rs
		(mu/Ypx)*cx]) #rp
	sym = [cx, cs, cp]
	return r, sym

#def rxRateAllNumpy(rc):
	#Computes the reaction rate vector for a given concentration, which is stored in the numpy array c


	#unpack the parameters in rc (rate constant)
	V_maxf = rc['V_maxf'] # mol/(g h)
	V_maxb = rc['V_maxb'] # mol/(g h)
	K_R1 = rc['K_R1'] # [mol/L]
	K_R2 = rc['K_R2'] # [mol/L]
	K_IR1 = rc['K_IR1'] # [mol/L]
	K_IR2 = rc['K_IR2'] # [mol/L]
	K_P1 = rc['K_P1'] # [mol/L]
	K_P2 = rc['K_P2'] # [mol/L]
	K_IP1 = rc['K_IP1'] # [mol/L]
	K_IP2 = rc['K_IP2'] # [mol/L]
	K_iR1 = rc['K_iR1'] # [mol/L]
	K_iP2 = rc['K_iP2'] # [mol/L]
	K_R1P2 = rc['K_R1P2'] # [mol/L]
	K_R2P1 = rc['K_R2P1'] # [mol/L]
	R1 = rc['ba'] # index in matrix
	R2 = rc['buoh'] # index in matrix
	P1 = rc['h2o'] # index in matrix
	P2 = rc['bb'] # index in matrix

	#The denominator in the rate expression
	D = V_maxb*c[R1]*c[R2] + V_maxb*K_R2*c[R1]*(1 + c[R1]/K_IR1) + V_maxb*K_R1*c[R2]*(1 + c[R2]/K_IR2) + (V_maxf*K_P1/K_eq)*c[P2]*(1 + c[P2]/K_IP2) + (V_maxf*K_P2/K_eq)*c[P1]*(1 + c[P1]/K_IP1) + (V_maxf/K_eq)*c[P1]*c[P2] + ((V_maxf*K_P2)/(K_eq*K_iR1))*c[R1]*c[P1] + ((V_maxb*K_R1)/K_iP2)*c[R1]*c[P2] + ((V_maxb*K_R1)/K_R2P1)*c[P1]*c[R2] + ((V_maxb*K_R2)/K_R1P2)*c[R1]*c[P2]

	#The rate vector, V
	V = (V_maxf*V_maxb*(c[R1]*c[R2] - (c[P1]*c[P2])/K_eq))/D

	r = np.zeros(np.shape(c))
	r[0] = V # dba/dt
	r[1] = V # dbuoh/dt
	r[2] = V # dh2o/dt
	r[3] = V # dbb/dt
	r[4] = 1 # dt/dt
	return r

#def rxRateAllSympy(rc):
	#Computes the reaction rate vector for a given concentration, which is stored in the numpy array c
	#unpack the parameters in rc (rate constant)
	muMax = rc['muMax']
	Ks = rc['Ks']
	kd = rc['kd']
	Yxs = rc['Yxs']
	Ypx = rc['Ypx']
	#Write the rate vector. In this case, rxRateAll=rxRate, however this is not general (this 
	# hopefully facilitates programming in a later stage)
	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
	mu = (muMax * cs) / (Ks + cs)
	r = Matrix([mu*cx - kd*cx, #rx
		-(mu/Yxs)*cx, #rs
		(mu/Ypx)*cx]) #rp

	# dr = r.jacobian([cb, ce, ct, cx, cd, ch])
	# rNumpyRaw = sympy.lambdify(('cb', 'ce', 'ct', 'cx', 'cd', 'ch'), r, 'numpy')
	# drNumpyRaw = sympy.lambdify(('cb', 'ce', 'ct', 'cx', 'cd', 'ch'), dr, 'numpy')
	# rNumpy = lambda c: rNumpyRaw(c[0], c[1], c[2], c[3], c[4], c[5])
	# drNumpy = lambda c: drNumpyRaw(c[0], c[1], c[2], c[3], c[4], c[5])
	return r