import numpy as np 
import sympy
from sympy.matrices import Matrix


def rxRateFun(c, rc):
	"""
	Computes the reaction rate vector for a given concentration

	INPUT
	c: numpy array. State space vector which the AR will be computed in (usually concentrations or
		concentrations and residence time space)
	rc: dictionary. Information about the rate constants
	OUTPUT
	r: numpy array. The rate vector
	"""
	# Unpack the rate vector
	B = rc['B']
	cf = rc['cf']
	cfAll = rc['cfAll']
	ix = rc['ix']
	isub = rc['isub']
	ihbu = rc['ihbu']# initial concentrations [g/L]
	ibuoh = rc['ibuoh']
	cs = c[0]
	cbuoh = c[1]
	r = np.zeros(np.shape(c))

	#Write the mass balance matrix. Have s and buoh as independent species
	#				s, buoh
	A = np.matrix([[-B[2], 0],#x
		[1, 0],#s
		[-B[3] - B[4]*B[9]/B[10], -B[4]/B[10]],#hbu
		[0, 1], #buoh
		[-B[13], 0]])#etoh
	# write delta-concentrations = c - cf for the independent species
	cfAll = np.asmatrix(cfAll[:-1]).T
	cf = np.asmatrix(cf).T
	c = np.asmatrix(c).T
	dci = c - cf #change of independent concentrations
	cAll = cfAll + A*dci #mass balance for dependent and independent species

	cx = cAll[ix, 0]
	chbu = cAll[ihbu, 0]

	rs = ((B[0] * cs) / (B[1] + cs))/B[2]
	
	r[0] = -rs*cx #dcs/dt
	r[1] = -B[9]*r[0] + B[10]*(chbu/(B[5] + chbu))*cx # dcbuoh/dt
	return r



def solveMoleBalance(cin, rc):
	"""
	Solve the mole/mass balance for the system (determine the dependent concentrations given the 
	independent concentrations)
	INPUT
	cin: numpy array. State space vector which the AR will be computed in (usually concentrations or
		concentrations and residence time space). The independent variables
	rc: dictionary. Information about the rate constants
	OUTPUT
	cout: numpy array. State space vector for all independent and dependent variabels (independent 
		concentrations + residence time (if this is variabel in the specific AR) AND dependent
		concentrations
	"""

	B = rc['B']
	cf = rc['cf']
	cfAll = rc['cfAll']
	ix = rc['ix']
	isub = rc['isub']
	ihbu = rc['ihbu']# initial concentrations [g/L]
	ibuoh = rc['ibuoh']
	ietoh = rc['ietoh']
	itau = rc['itau']

	#Write the mass balance matrix. Have s and buoh as independent species
	#				s, buoh
	A = np.matrix([[-B[2], 0],#x
		[1, 0],#s
		[-B[3] - B[4]*B[9]/B[10], -B[4]/B[10]],#hbu
		[0, 1], #buoh
		[-B[13], 0]])#etoh
	# write delta-concentrations = c - cf for the independent species
	cfAll = np.asmatrix(cfAll[:-1]).T
	cf = np.asmatrix(cf).T
	cAll = np.zeros((cin.shape[0], len(cfAll)))
	for i in range(len(cin[:, 0])):
		cInd = np.asmatrix(cin[i, :]).T
		# ci = np.matrix([[cs], [cbuoh]])
		dci = cInd - cf #change of independent concentrations
		cAlli = cfAll + A*dci #mass balance for dependent and independent species
		cAll[i, :] = np.asarray(cAlli)[:, 0]
	cout = np.column_stack((cAll[:, ix], cAll[:, isub], cAll[:, ihbu], cAll[:, ibuoh], cAll[:, ietoh]))
	return cout


# def rxRateAllNumpy(rc):
# 	#Computes the reaction rate vector for a given concentration, which is stored in the numpy array c
# 	#unpack the parameters in rc (rate constant)
# 	muMax = rc['muMax']
# 	Ks = rc['Ks']
# 	kd = rc['kd']
# 	Yxs = rc['Yxs']
# 	Ypx = rc['Ypx']
# 	cx0 = rc['cx0']# initial concentrations [g/L]
# 	cs0 = rc['cs0']
# 	cBuOH0 = rc['cBuOH0']
# 	#Write the rate vector. In this case, rxRateAll=rxRate, however this is not general (this 
# 	# hopefully facilitates programming in a later stage)
# 	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
# 	mu = (muMax * cs) / (Ks + cs)
# 	r = Matrix([mu*cx - kd*cx, #rx
# 		-(mu/Yxs)*cx, #rs
# 		(mu/Ypx)*cx]) #rp

# 	dr = r.jacobian([cx, cs, cp])
# 	rNumpyRaw = sympy.lambdify(('cx', 'cs', 'cp'), r, 'numpy')
# 	drNumpyRaw = sympy.lambdify(('cx', 'cs', 'cp'), r, 'numpy')
# 	rNumpy = lambda c: rNumpyRaw(c[0], c[1], c[2])
# 	drNumpy = lambda c: drNumpyRaw(c[0], c[1], c[2])
# 	return rNumpy, drNumpy


# The following kinetics are not updated (written for cBuOH instead of cS as output)
# def rxRateSympy(rc):
# 	"""
# 	Returns the reaction rate vector with symbolic expressions. Also return the expressions themselves
# 	INPUT
# 	rc: dictionary. Information about the rate constants
# 	OUTPUT
# 	r: sympy Matrix. The rate vector with symbolic expressions
# 	sym: list of sympy.symbols
# 	"""
# 	muMax = rc['muMax']
# 	Ks = rc['Ks']
# 	kd = rc['kd']
# 	Yxs = rc['Yxs']
# 	Ypx = rc['Ypx']
# 	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
# 	mu = (muMax * cs) / (Ks + cs)
# 	r = Matrix([mu*cx - kd*cx, #rx
# 		-(mu/Yxs)*cx, #rs
# 		(mu/Ypx)*cx]) #rp
# 	sym = [cx, cs, cp]
# 	return r, sym



# def rxRateAllSympy(rc):
# 	#Computes the reaction rate vector for a given concentration, which is stored in the numpy array c
# 	#unpack the parameters in rc (rate constant)
# 	muMax = rc['muMax']
# 	Ks = rc['Ks']
# 	kd = rc['kd']
# 	Yxs = rc['Yxs']
# 	Ypx = rc['Ypx']
# 	#Write the rate vector. In this case, rxRateAll=rxRate, however this is not general (this 
# 	# hopefully facilitates programming in a later stage)
# 	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
# 	mu = (muMax * cs) / (Ks + cs)
# 	r = Matrix([mu*cx - kd*cx, #rx
# 		-(mu/Yxs)*cx, #rs
# 		(mu/Ypx)*cx]) #rp

# 	# dr = r.jacobian([cb, ce, ct, cx, cd, ch])
# 	# rNumpyRaw = sympy.lambdify(('cb', 'ce', 'ct', 'cx', 'cd', 'ch'), r, 'numpy')
# 	# drNumpyRaw = sympy.lambdify(('cb', 'ce', 'ct', 'cx', 'cd', 'ch'), dr, 'numpy')
# 	# rNumpy = lambda c: rNumpyRaw(c[0], c[1], c[2], c[3], c[4], c[5])
# 	# drNumpy = lambda c: drNumpyRaw(c[0], c[1], c[2], c[3], c[4], c[5])
# 	return r