import numpy as np 
import sympy
from sympy.matrices import Matrix
def rxRateFun(c, rc):
	"""
	Computes the reaction rate vector for a given concentration

	INPUT
	c: numpy array. State space vector which the AR will be computed in (usually concentrations or
		concentrations and residence time space)
	rc: dictionary. Information about the rate constants
	OUTPUT
	r: numpy array. The rate vector
	"""

	# Unpack the rate vector. For units, check the definition of rc in the file "main.py"
	mu_m = rc['mu_m']
	exponent = rc['exponent']
	m_S = rc['m_S']
	K_I = rc['K_I']
	K_d = rc['K_d']
	K_S = rc['K_S']
	P_d = rc['P_d']
	alpha_Aa = rc['alpha_Aa']
	alpha_Ba = rc['alpha_Ba']
	beta_Aa = rc['beta_Aa']
	beta_Ba = rc['beta_Ba']
	Y_X = rc['Y_X']
	Y_Aa = rc['Y_Aa']
	Y_Ba = rc['Y_Ba']
	x = rc['x_ind']# initial concentrations [g/L]
	s = rc['s_ind']
	aa = rc['aa_ind']
	ba = rc['ba_ind']
	tau_ind = rc['tau_ind']
	cf = rc['cfAll'] #containes the feed concentration of all the components in the system
	N = rc['N']
	MW = rc['MW']

	#unpack the concentration vector
	cs = c[0]
	tau = c[1]

	#simplify notation, "invent" a variable for dsdt
	alpha_S = 1/Y_X + (alpha_Aa/Y_Aa)*(1 + 1*44.01/60.05) + (alpha_Ba/Y_Ba)*(1 + 2*44.01/88.1)

	#For simplification of the rate expression, compute cx, cs and caa
	#(could've substituted inside the function)
	cx = cf[x] - (1/alpha_S) * (cs - cf[s]) #from dxdt = alpha_ba*dba/dt
	cba = cf[ba] - (alpha_Ba/alpha_S) * (cs - cf[s]) # from dsdt = -(alphaS/alphaBa)*dba/dt
	caa = cf[aa] - (alpha_Aa/alpha_S) * (cs - cf[s])
	# print(cx*MW[x],cs*MW[s],caa*MW[aa], cba*MW[ba], tau)
	#compute the rate vector
	mu = ((mu_m * cs) / (cs + K_S + 
		((cs**2)/K_I))) * ((1 - (cba + caa)/P_d)**exponent)
	r = np.zeros(np.shape(c))
	r[0] = -alpha_S*(mu*cx) # [g/(L h)], dcs/dt
	r[1] = 1 #dt/dt
	return r

def solveMoleBalance(cin, rc):
	"""
	Solve the mole/mass balance for the system (determine the dependent concentrations given the 
	independent concentrations)
	INPUT
	cin: numpy array. State space vector which the AR will be computed in (usually concentrations or
		concentrations and residence time space). The independent variables
	rc: dictionary. Information about the rate constants
	OUTPUT
	cout: numpy array. State space vector for all independent and dependent variabels (independent 
		concentrations + residence time (if this is variabel in the specific AR) AND dependent
		concentrations

	NOTE: cin IS NOW IN [G/L], NEED TO CONVERT TO MOLES (should probably be done in main file)
	"""

	# Unpack the rate vector
	mu_m = rc['mu_m']
	exponent = rc['exponent']
	m_S = rc['m_S']
	K_I = rc['K_I']
	K_d = rc['K_d']
	K_S = rc['K_S']
	P_d = rc['P_d']
	alpha_Aa = rc['alpha_Aa']
	alpha_Ba = rc['alpha_Ba']
	beta_Aa = rc['beta_Aa']
	beta_Ba = rc['beta_Ba']
	Y_X = rc['Y_X']
	Y_Aa = rc['Y_Aa']
	Y_Ba = rc['Y_Ba']
	x = rc['x_ind']# initial concentrations [g/L]
	s = rc['s_ind']
	aa = rc['aa_ind']
	ba = rc['ba_ind']
	tau_ind = rc['tau_ind']
	cf = rc['cfAll'] #containes the feed concentration of all the components in the system
	N = rc['N']
	MW = rc['MW']

	#unpack the concentration vector
	cs = cin[:, 0]
	tau = cin[:, 1]

	#simplify notation, "invent" a variable for dsdt
	alpha_S = 1/Y_X + (alpha_Aa/Y_Aa)*(1 + 1*44.01/60.05) + (alpha_Ba/Y_Ba)*(1 + 2*44.01/88.1)

	#For simplification of the rate expression, compute cx, cs and caa
	#(could've substituted inside the function)
	cx = cf[x] - (1/alpha_S) * (cs - cf[s]) #from dxdt = alpha_ba*dba/dt
	cba = cf[ba] - (alpha_Ba/alpha_S) * (cs - cf[s]) # from dsdt = -(alphaS/alphaBa)*dba/dt
	caa = cf[aa] - (alpha_Aa/alpha_S) * (cs - cf[s])
	#stack the variables in the output concentrations vector
	cout = np.column_stack((cs, cx, caa, cba, tau))
	# cout = cout[~np.isnan(cout).any(axis=1)] #removes the rows which contains nan

	return cout


# #def rxRateSympy(rc):

# 	"""
# 	Returns the reaction rate vector with symbolic expressions. Also return the expressions themselves
# 	INPUT
# 	rc: dictionary. Information about the rate constants
# 	OUTPUT
# 	r: sympy Matrix. The rate vector with symbolic expressions
# 	sym: list of sympy.symbols
# 	"""

# 	muMax = rc['muMax']
# 	Ks = rc['Ks']
# 	kd = rc['kd']
# 	Yxs = rc['Yxs']
# 	Ypx = rc['Ypx']
# 	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
# 	mu = (muMax * cs) / (Ks + cs)
# 	r = Matrix([mu*cx - kd*cx, #rx
# 		-(mu/Yxs)*cx, #rs
# 		(mu/Ypx)*cx]) #rp
# 	sym = [cx, cs, cp]
# 	return r, sym

# #def rxRateAllNumpy(rc):
# 	#Computes the reaction rate vector for a given concentration, which is stored in the numpy array c
# 	#unpack the parameters in rc (rate constant)
# 	muMax = rc['muMax']
# 	Ks = rc['Ks']
# 	kd = rc['kd']
# 	Yxs = rc['Yxs']
# 	Ypx = rc['Ypx']
# 	#Write the rate vector. In this case, rxRateAll=rxRate, however this is not general (this 
# 	# hopefully facilitates programming in a later stage)
# 	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
# 	mu = (muMax * cs) / (Ks + cs)
# 	r = Matrix([mu*cx - kd*cx, #rx
# 		-(mu/Yxs)*cx, #rs
# 		(mu/Ypx)*cx]) #rp

# 	dr = r.jacobian([cx, cs, cp])
# 	rNumpyRaw = sympy.lambdify(('cx', 'cs', 'cp'), r, 'numpy')
# 	drNumpyRaw = sympy.lambdify(('cx', 'cs', 'cp'), r, 'numpy')
# 	rNumpy = lambda c: rNumpyRaw(c[0], c[1], c[2])
# 	drNumpy = lambda c: drNumpyRaw(c[0], c[1], c[2])
# 	return rNumpy, drNumpy


# #def rxRateAllSympy(rc):
# 	#Computes the reaction rate vector for a given concentration, which is stored in the numpy array c
# 	#unpack the parameters in rc (rate constant)
# 	muMax = rc['muMax']
# 	Ks = rc['Ks']
# 	kd = rc['kd']
# 	Yxs = rc['Yxs']
# 	Ypx = rc['Ypx']
# 	#Write the rate vector. In this case, rxRateAll=rxRate, however this is not general (this 
# 	# hopefully facilitates programming in a later stage)
# 	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
# 	mu = (muMax * cs) / (Ks + cs)
# 	r = Matrix([mu*cx - kd*cx, #rx
# 		-(mu/Yxs)*cx, #rs
# 		(mu/Ypx)*cx]) #rp

# 	# dr = r.jacobian([cb, ce, ct, cx, cd, ch])
# 	# rNumpyRaw = sympy.lambdify(('cb', 'ce', 'ct', 'cx', 'cd', 'ch'), r, 'numpy')
# 	# drNumpyRaw = sympy.lambdify(('cb', 'ce', 'ct', 'cx', 'cd', 'ch'), dr, 'numpy')
# 	# rNumpy = lambda c: rNumpyRaw(c[0], c[1], c[2], c[3], c[4], c[5])
# 	# drNumpy = lambda c: drNumpyRaw(c[0], c[1], c[2], c[3], c[4], c[5])
# 	return r
