import numpy as np 
import sympy
from sympy.matrices import Matrix
def rxRateFun(c, rc):
	"""
	Computes the reaction rate vector for a given concentration

	INPUT
	c: numpy array. State space vector which the AR will be computed in (usually concentrations or
		concentrations and residence time space)
	rc: dictionary. Information about the rate constants
	OUTPUT
	r: numpy array. The rate vector
	"""

	# Unpack the rate vector. For units, check the definition of rc in the file "main.py"
	phbu = rc['phbu'] # parameters for hbu model
	pbub = rc['pbub'] # parameters for bub model

	# Unpack hbu model
	mu_m = phbu['mu_m']
	exponent = phbu['exponent']
	m_S = phbu['m_S']
	K_I = phbu['K_I']
	K_d = phbu['K_d']
	K_S = phbu['K_S']
	P_d = phbu['P_d']
	alpha_Aa = phbu['alpha_Aa']
	alpha_Ba = phbu['alpha_Ba']
	beta_Aa = phbu['beta_Aa']
	beta_Ba = phbu['beta_Ba']
	Y_X = phbu['Y_X']
	Y_Aa = phbu['Y_Aa']
	Y_Ba = phbu['Y_Ba']

	# Unpack bub model
	rmax = pbub['rmax'] # mol/(L h)
	Kma = pbub['Kma'] # [mol/L]
	Kmb = pbub['Kmb'] # [mol/L]
	Kia = pbub['Kia'] # [mol/L]

	# Unpack rest of the parameters
	cfAll = np.asmatrix(rc['cfAll']).T #containes the feed concentration of all the components in the system
	cf = np.asmatrix(rc['cf']).T
	MW = rc['MW']
	isub = rc['isub'] # indices in vectors
	ihbu = rc['ihbu']
	ibuoh = rc['ibuoh']
	ibub = rc['ibub']
	ix = rc['ix']
	ihac = rc['ihac']

	# Make the independent concentrations in matrix form
	ci = np.asmatrix(c).T #independt species: S, HBu, BuB

	# Write the (combined) kinetic model
	# First; simplify notation. "Invent" helping variables gamma and omega
	gamma = 1/Y_X + (alpha_Aa/Y_Aa)*(1 + 1*44.01/60.05) + (alpha_Ba/Y_Ba)*(1 + 2*44.01/88.1)
	omega = (1/Y_Ba)*(1 + 2*44.01/88.1)

	# Make the reaction rate matrix for all the components
	A = np.matrix([[-1/gamma, (omega/gamma) * (MW[ihbu]/MW[ibub])],# dcX/dt
		[1, 0], #dcS/dt
		[-alpha_Aa/gamma, alpha_Aa * (omega/gamma) * (MW[ihbu]/MW[ibub])], # dcHac/dt
		[-alpha_Ba/gamma, (MW[ihbu]/MW[ibub]) * (alpha_Ba * (omega/gamma) - 1)], #dcHBu/dt
		[0, -MW[ibuoh]/MW[ibub]], # dcBuOH/dt
		[0, 1]]) # dcBub/dt 

	# Compute the concentration of all components
	cAll = cfAll + A*(ci - cf)


	# For the Bub model: compute the rate of reaction, in molar form. Rewrite chbu = ca and 
	# cbuoh = cb, where ca,cb are in mol/L
	ca = cAll[ihbu, 0]/MW[ihbu]
	cb = cAll[ibuoh, 0]/MW[ibuoh]

	rateOfRxBubModelMolar = (rmax * ca * cb) / (Kia*Kmb + Kma*cb + Kmb*ca + ca*cb) #[mol/(L h)]

	# Calculate the cell growth rate (HBu-model)
	mu = ((mu_m * cAll[isub, 0]) / (cAll[isub, 0] + K_S + 
		((cAll[isub, 0]**2)/K_I))) * ((1 - (cAll[ihbu, 0] + cAll[ihac, 0])/P_d)**exponent)

	# Get the rate vector for the independent species
	r = np.zeros(np.shape(c))
	r[0] = -gamma*(mu*cAll[ix, 0]) + omega * MW[ihbu] * rateOfRxBubModelMolar # dcS/dt, [g/(L h)]
	r[1] = MW[ibub] * rateOfRxBubModelMolar # dcBub/dt [g/(L h)]
	return r

def solveMoleBalance(cin, rc):
	"""
	Solve the mole/mass balance for the system (determine the dependent concentrations given the 
	independent concentrations)
	INPUT
	cin: numpy array. State space vector which the AR will be computed in (usually concentrations or
		concentrations and residence time space). The independent variables
	rc: dictionary. Information about the rate constants
	OUTPUT
	cout: numpy array. State space vector for all independent and dependent variabels (independent 
		concentrations + residence time (if this is variabel in the specific AR) AND dependent
		concentrations

	NOTE: cin IS NOW IN [G/L], NEED TO CONVERT TO MOLES (should probably be done in main file)
	"""

	# Unpack the rate vector
	phbu = rc['phbu'] # parameters for hbu model
	pbub = rc['pbub'] # parameters for bub model

	# Unpack hbu model
	mu_m = phbu['mu_m']
	exponent = phbu['exponent']
	m_S = phbu['m_S']
	K_I = phbu['K_I']
	K_d = phbu['K_d']
	K_S = phbu['K_S']
	P_d = phbu['P_d']
	alpha_Aa = phbu['alpha_Aa']
	alpha_Ba = phbu['alpha_Ba']
	beta_Aa = phbu['beta_Aa']
	beta_Ba = phbu['beta_Ba']
	Y_X = phbu['Y_X']
	Y_Aa = phbu['Y_Aa']
	Y_Ba = phbu['Y_Ba']

	# Unpack bub model
	rmax = pbub['rmax'] # mol/(L h)
	Kma = pbub['Kma'] # [mol/L]
	Kmb = pbub['Kmb'] # [mol/L]
	Kia = pbub['Kia'] # [mol/L]

	# Unpack rest of the parameters
	cfAll = np.asmatrix(rc['cfAll']).T #containes the feed concentration of all the components in the system
	cf = np.asmatrix(rc['cf']).T
	MW = rc['MW']
	isub = rc['isub'] # indices in vectors
	ihbu = rc['ihbu']
	ibuoh = rc['ibuoh']
	ibub = rc['ibub']
	ix = rc['ix']
	ihac = rc['ihac']

	# Write the (combined) kinetic model
	# First; simplify notation. "Invent" helping variables gamma and omega
	gamma = 1/Y_X + (alpha_Aa/Y_Aa)*(1 + 1*44.01/60.05) + (alpha_Ba/Y_Ba)*(1 + 2*44.01/88.1)
	omega = (1/Y_Ba)*(1 + 2*44.01/88.1)

	# Make the reaction rate matrix for all the components
	A = np.matrix([[-1/gamma, (omega/gamma) * (MW[ihbu]/MW[ibub])],# dcX/dt
		[1, 0], #dcS/dt
		[-alpha_Aa/gamma, alpha_Aa * (omega/gamma) * (MW[ihbu]/MW[ibub])], # dcHac/dt
		[-alpha_Ba/gamma, (MW[ihbu]/MW[ibub]) * (alpha_Ba * (omega/gamma) - 1)], #dcHBu/dt
		[0, -MW[ibuoh]/MW[ibub]], # dcBuOH/dt
		[0, 1]]) # dcBub/dt 

	cAll = np.zeros((cin.shape[0], len(cfAll)))
	for i in range(len(cin[:, 0])):
		cInd = np.asmatrix(cin[i, :]).T
		# ci = np.matrix([[cs], [cbuoh]])
		dci = cInd - cf #change of independent concentrations
		cAlli = cfAll + A*dci #mass balance for dependent and independent species
		cAll[i, :] = np.asarray(cAlli)[:, 0]
	# cout = np.column_stack((cAll[:, ix], cAll[:, isub], cAll[:, ihbu], cAll[:, ibuoh], cAll[:, ietoh]))
	return cAll

def solveMoleBalancePointStateSpace(cin, rc):
	"""
	Solve the mole/mass balance for a single point in the state fspace for the system 
	(determine the dependent concentrations given the independent concentrations)
	INPUT
	cin: 1D numpy array. State space vector which the AR will be computed in (usually concentrations or
		concentrations and residence time space). The independent variables.
	rc: dictionary. Information about the rate constants
	OUTPUT
	cout: 1D numpy array. State space vector for all independent and dependent variabels (independent 
		concentrations + residence time (if this is variabel in the specific AR) AND dependent
		concentrations

	"""

	# Unpack the rate vector
	phbu = rc['phbu'] # parameters for hbu model
	pbub = rc['pbub'] # parameters for bub model

	# Unpack hbu model
	mu_m = phbu['mu_m']
	exponent = phbu['exponent']
	m_S = phbu['m_S']
	K_I = phbu['K_I']
	K_d = phbu['K_d']
	K_S = phbu['K_S']
	P_d = phbu['P_d']
	alpha_Aa = phbu['alpha_Aa']
	alpha_Ba = phbu['alpha_Ba']
	beta_Aa = phbu['beta_Aa']
	beta_Ba = phbu['beta_Ba']
	Y_X = phbu['Y_X']
	Y_Aa = phbu['Y_Aa']
	Y_Ba = phbu['Y_Ba']

	# Unpack bub model
	rmax = pbub['rmax'] # mol/(L h)
	Kma = pbub['Kma'] # [mol/L]
	Kmb = pbub['Kmb'] # [mol/L]
	Kia = pbub['Kia'] # [mol/L]

	# Unpack rest of the parameters
	cfAll = np.asmatrix(rc['cfAll']).T #containes the feed concentration of all the components in the system
	cf = np.asmatrix(rc['cf']).T
	MW = rc['MW']
	isub = rc['isub'] # indices in vectors
	ihbu = rc['ihbu']
	ibuoh = rc['ibuoh']
	ibub = rc['ibub']
	ix = rc['ix']
	ihac = rc['ihac']

	# Write the (combined) kinetic model
	# First; simplify notation. "Invent" helping variables gamma and omega
	gamma = 1/Y_X + (alpha_Aa/Y_Aa)*(1 + 1*44.01/60.05) + (alpha_Ba/Y_Ba)*(1 + 2*44.01/88.1)
	omega = (1/Y_Ba)*(1 + 2*44.01/88.1)

	# Make the reaction rate matrix for all the components
	A = np.matrix([[-1/gamma, (omega/gamma) * (MW[ihbu]/MW[ibub])],# dcX/dt
		[1, 0], #dcS/dt
		[-alpha_Aa/gamma, alpha_Aa * (omega/gamma) * (MW[ihbu]/MW[ibub])], # dcHac/dt
		[-alpha_Ba/gamma, (MW[ihbu]/MW[ibub]) * (alpha_Ba * (omega/gamma) - 1)], #dcHBu/dt
		[0, -MW[ibuoh]/MW[ibub]], # dcBuOH/dt
		[0, 1]]) # dcBub/dt 

	ci = np.asmatrix(cin).T

	cAll = cfAll + A*(ci - cf)
	cout = np.asarray(cAll.T)[0, :]

	return cout

# #def rxRateSympy(rc):

# 	"""
# 	Returns the reaction rate vector with symbolic expressions. Also return the expressions themselves
# 	INPUT
# 	rc: dictionary. Information about the rate constants
# 	OUTPUT
# 	r: sympy Matrix. The rate vector with symbolic expressions
# 	sym: list of sympy.symbols
# 	"""

# 	muMax = rc['muMax']
# 	Ks = rc['Ks']
# 	kd = rc['kd']
# 	Yxs = rc['Yxs']
# 	Ypx = rc['Ypx']
# 	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
# 	mu = (muMax * cs) / (Ks + cs)
# 	r = Matrix([mu*cx - kd*cx, #rx
# 		-(mu/Yxs)*cx, #rs
# 		(mu/Ypx)*cx]) #rp
# 	sym = [cx, cs, cp]
# 	return r, sym

# #def rxRateAllNumpy(rc):
# 	#Computes the reaction rate vector for a given concentration, which is stored in the numpy array c
# 	#unpack the parameters in rc (rate constant)
# 	muMax = rc['muMax']
# 	Ks = rc['Ks']
# 	kd = rc['kd']
# 	Yxs = rc['Yxs']
# 	Ypx = rc['Ypx']
# 	#Write the rate vector. In this case, rxRateAll=rxRate, however this is not general (this 
# 	# hopefully facilitates programming in a later stage)
# 	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
# 	mu = (muMax * cs) / (Ks + cs)
# 	r = Matrix([mu*cx - kd*cx, #rx
# 		-(mu/Yxs)*cx, #rs
# 		(mu/Ypx)*cx]) #rp

# 	dr = r.jacobian([cx, cs, cp])
# 	rNumpyRaw = sympy.lambdify(('cx', 'cs', 'cp'), r, 'numpy')
# 	drNumpyRaw = sympy.lambdify(('cx', 'cs', 'cp'), r, 'numpy')
# 	rNumpy = lambda c: rNumpyRaw(c[0], c[1], c[2])
# 	drNumpy = lambda c: drNumpyRaw(c[0], c[1], c[2])
# 	return rNumpy, drNumpy


# #def rxRateAllSympy(rc):
# 	#Computes the reaction rate vector for a given concentration, which is stored in the numpy array c
# 	#unpack the parameters in rc (rate constant)
# 	muMax = rc['muMax']
# 	Ks = rc['Ks']
# 	kd = rc['kd']
# 	Yxs = rc['Yxs']
# 	Ypx = rc['Ypx']
# 	#Write the rate vector. In this case, rxRateAll=rxRate, however this is not general (this 
# 	# hopefully facilitates programming in a later stage)
# 	cx, cs, cp = sympy.symbols('cx,cs,cp', real=True) 
# 	mu = (muMax * cs) / (Ks + cs)
# 	r = Matrix([mu*cx - kd*cx, #rx
# 		-(mu/Yxs)*cx, #rs
# 		(mu/Ypx)*cx]) #rp

# 	# dr = r.jacobian([cb, ce, ct, cx, cd, ch])
# 	# rNumpyRaw = sympy.lambdify(('cb', 'ce', 'ct', 'cx', 'cd', 'ch'), r, 'numpy')
# 	# drNumpyRaw = sympy.lambdify(('cb', 'ce', 'ct', 'cx', 'cd', 'ch'), dr, 'numpy')
# 	# rNumpy = lambda c: rNumpyRaw(c[0], c[1], c[2], c[3], c[4], c[5])
# 	# drNumpy = lambda c: drNumpyRaw(c[0], c[1], c[2], c[3], c[4], c[5])
# 	return r
